@@ -28,20 +28,17 @@ TODO:
 
 import os
 import sys
+import warnings
 
 from typing import BinaryIO, Tuple, Optional
 
-from collections import (
-    OrderedDict,
-)
-
 try:
     from collections.abc import (
         Iterable,
         MutableMapping,
     )
 except ImportError:  # python < 3.7
-    from collections import (
+    from collections import (  # type: ignore
         Iterable,
         MutableMapping,
     )
@@ -62,7 +59,12 @@ def lower_key(key):
     return key
 
 
-class CaseInsensitiveDict(OrderedDict):
+class CaseInsensitiveOrderedMultiDict(MutableMapping):
+
+    def __init__(self):
+        self._real = []
+        self._keyed = {}
+
     @classmethod
     def make(cls, dict_in=None):
 
@@ -82,15 +84,34 @@ class CaseInsensitiveDict(OrderedDict):
 
         return out
 
-    def __setitem__(self, key, value, **kwargs):
-        key = lower_key(key)
+    def __len__(self):
+        return len(self._keyed)
 
-        super(CaseInsensitiveDict, self).__setitem__(key, value, **kwargs)
+    def keys(self):
+        return self._keyed.keys()
+
+    def items(self):
+        return iter(self._real)
+
+    def __iter__(self):
+        return self._keyed.__iter__()
+
+    def values(self):
+        return self._keyed.values()
+
+    def __setitem__(self, key, value):
+        self._real.append((key, value))
+        self._keyed[lower_key(key)] = value
+
+    def __delitem__(self, key):
+        key = lower_key(key)
+        del self._keyed[key]
+        for i, (actual, unused_value) in reversed(list(enumerate(self._real))):
+            if lower_key(actual) == key:
+                del self._real[i]
 
     def __getitem__(self, item):
-        key = lower_key(item)
-
-        return super(CaseInsensitiveDict, self).__getitem__(key)
+        return self._keyed[lower_key(item)]
 
     def get(self, key, default=SENTINAL):
         try:
@@ -103,6 +124,12 @@ class CaseInsensitiveDict(OrderedDict):
 
         return default
 
+    def get_all(self, key):
+        key = lower_key(key)
+        for actual, value in self._real:
+            if lower_key(actual) == key:
+                yield value
+
     def setdefault(self, key, default=SENTINAL):
         try:
             return self[key]
@@ -120,7 +147,7 @@ class Config(object):
 
         Args:
           section: Tuple with section name and optional subsection namee
-          subsection: Subsection name
+          name: Variable name
         Returns:
           Contents of the setting
         Raises:
@@ -128,6 +155,19 @@ class Config(object):
         """
         raise NotImplementedError(self.get)
 
+    def get_multivar(self, section, name):
+        """Retrieve the contents of a multivar configuration setting.
+
+        Args:
+          section: Tuple with section name and optional subsection namee
+          name: Variable name
+        Returns:
+          Contents of the setting as iterable
+        Raises:
+          KeyError: if the value is not set
+        """
+        raise NotImplementedError(self.get_multivar)
+
     def get_boolean(self, section, name, default=None):
         """Retrieve a configuration setting as boolean.
 
@@ -157,10 +197,20 @@ class Config(object):
           section: Tuple with section name and optional subsection namee
           name: Name of the configuration value, including section
             and optional subsection
-           value: value of the setting
+          value: value of the setting
         """
         raise NotImplementedError(self.set)
 
+    def items(self, section):
+        """Iterate over the configuration pairs for a specific section.
+
+        Args:
+          section: Tuple with section name and optional subsection namee
+        Returns:
+          Iterator over (name, value) pairs
+        """
+        raise NotImplementedError(self.items)
+
     def iteritems(self, section):
         """Iterate over the configuration pairs for a specific section.
 
@@ -169,14 +219,27 @@ class Config(object):
         Returns:
           Iterator over (name, value) pairs
         """
-        raise NotImplementedError(self.iteritems)
+        warnings.warn(
+            "Use %s.items instead." % type(self).__name__,
+            DeprecationWarning,
+            stacklevel=3,
+        )
+        return self.items(section)
 
     def itersections(self):
+        warnings.warn(
+            "Use %s.items instead." % type(self).__name__,
+            DeprecationWarning,
+            stacklevel=3,
+        )
+        return self.sections()
+
+    def sections(self):
         """Iterate over the sections.
 
         Returns: Iterator over section tuples
         """
-        raise NotImplementedError(self.itersections)
+        raise NotImplementedError(self.sections)
 
     def has_section(self, name):
         """Check if a specified section exists.
@@ -186,7 +249,7 @@ class Config(object):
         Returns:
           boolean indicating whether the section exists
         """
-        return name in self.itersections()
+        return name in self.sections()
 
 
 class ConfigDict(Config, MutableMapping):
@@ -197,7 +260,7 @@ class ConfigDict(Config, MutableMapping):
         if encoding is None:
             encoding = sys.getdefaultencoding()
         self.encoding = encoding
-        self._values = CaseInsensitiveDict.make(values)
+        self._values = CaseInsensitiveOrderedMultiDict.make(values)
 
     def __repr__(self):
         return "%s(%r)" % (self.__class__.__name__, self._values)
@@ -246,6 +309,17 @@ class ConfigDict(Config, MutableMapping):
 
         return section, name
 
+    def get_multivar(self, section, name):
+        section, name = self._check_section_and_name(section, name)
+
+        if len(section) > 1:
+            try:
+                return self._values[section][name]
+            except KeyError:
+                pass
+
+        return self._values[(section[0],)].get_all(name)
+
     def get(self, section, name):
         section, name = self._check_section_and_name(section, name)
 
@@ -265,10 +339,10 @@ class ConfigDict(Config, MutableMapping):
 
         self._values.setdefault(section)[name] = value
 
-    def iteritems(self, section):
+    def items(self, section):
         return self._values.get(section).items()
 
-    def itersections(self):
+    def sections(self):
         return self._values.keys()
 
 
@@ -387,14 +461,16 @@ class ConfigFile(ConfigDict):
         super(ConfigFile, self).__init__(values=values, encoding=encoding)
         self.path = None
 
-    @classmethod
-    def from_file(cls, f: BinaryIO) -> "ConfigFile":
+    @classmethod  # noqa: C901
+    def from_file(cls, f: BinaryIO) -> "ConfigFile":  # noqa: C901
         """Read configuration from a file-like object."""
         ret = cls()
         section = None  # type: Optional[Tuple[bytes, ...]]
         setting = None
         continuation = None
         for lineno, line in enumerate(f.readlines()):
+            if lineno == 0 and line.startswith(b'\xef\xbb\xbf'):
+                line = line[3:]
             line = line.lstrip()
             if setting is None:
                 # Parse section header ("[bla]")
