@@ -30,7 +30,6 @@ import stat
 import struct
 import sys
 import tempfile
-import warnings
 
 from dulwich.index import (
     Index,
@@ -48,6 +47,7 @@ from dulwich.index import (
     write_index_dict,
     _tree_to_fs_path,
     _fs_to_tree_path,
+    IndexEntry,
 )
 from dulwich.object_store import (
     MemoryObjectStore,
@@ -63,9 +63,6 @@ from dulwich.tests import (
     TestCase,
     skipIf,
 )
-from dulwich.tests.utils import (
-    setup_warning_catcher,
-)
 
 
 def can_symlink():
@@ -108,26 +105,6 @@ class SimpleIndexTestCase(IndexTestCase):
             list(self.get_simple_index("index").iterobjects()),
         )
 
-    def test_iterblobs(self):
-        warnings.simplefilter("always", UserWarning)
-        self.addCleanup(warnings.resetwarnings)
-        warnings_list, restore_warnings = setup_warning_catcher()
-        self.addCleanup(restore_warnings)
-
-        self.assertEqual(
-            [(b"bla", b"e69de29bb2d1d6434b8b29ae775ad8c2e48c5391", 33188)],
-            list(self.get_simple_index("index").iterblobs()),
-        )
-
-        expected_warning = PendingDeprecationWarning("Use iterobjects() instead.")
-        for w in warnings_list:
-            if type(w) == type(expected_warning) and w.args == expected_warning.args:
-                break
-        else:
-            raise AssertionError(
-                "Expected warning %r not in %r" % (expected_warning, warnings_list)
-            )
-
     def test_getitem(self):
         self.assertEqual(
             (
@@ -141,6 +118,7 @@ class SimpleIndexTestCase(IndexTestCase):
                 0,
                 b"e69de29bb2d1d6434b8b29ae775ad8c2e48c5391",
                 0,
+                0,
             ),
             self.get_simple_index("index")[b"bla"],
         )
@@ -172,16 +150,18 @@ class SimpleIndexWriterTestCase(IndexTestCase):
         entries = [
             (
                 b"barbla",
-                (1230680220, 0),
-                (1230680220, 0),
-                2050,
-                3761020,
-                33188,
-                1000,
-                1000,
-                0,
-                b"e69de29bb2d1d6434b8b29ae775ad8c2e48c5391",
-                0,
+                IndexEntry(
+                    (1230680220, 0),
+                    (1230680220, 0),
+                    2050,
+                    3761020,
+                    33188,
+                    1000,
+                    1000,
+                    0,
+                    b"e69de29bb2d1d6434b8b29ae775ad8c2e48c5391",
+                    0,
+                    0)
             )
         ]
         filename = os.path.join(self.tempdir, "test-simple-write-index")
@@ -203,7 +183,7 @@ class ReadIndexDictTests(IndexTestCase):
 
     def test_simple_write(self):
         entries = {
-            b"barbla": (
+            b"barbla": IndexEntry(
                 (1230680220, 0),
                 (1230680220, 0),
                 2050,
@@ -214,6 +194,7 @@ class ReadIndexDictTests(IndexTestCase):
                 0,
                 b"e69de29bb2d1d6434b8b29ae775ad8c2e48c5391",
                 0,
+                0,
             )
         }
         filename = os.path.join(self.tempdir, "test-simple-write-index")
@@ -314,7 +295,7 @@ class IndexEntryFromStatTests(TestCase):
         entry = index_entry_from_stat(st, "22" * 20, 0)
         self.assertEqual(
             entry,
-            (
+            IndexEntry(
                 1324180496,
                 1324180496,
                 64769,
@@ -325,6 +306,7 @@ class IndexEntryFromStatTests(TestCase):
                 12288,
                 "2222222222222222222222222222222222222222",
                 0,
+                None,
             ),
         )
 
@@ -346,7 +328,7 @@ class IndexEntryFromStatTests(TestCase):
         entry = index_entry_from_stat(st, "22" * 20, 0, mode=stat.S_IFREG + 0o755)
         self.assertEqual(
             entry,
-            (
+            IndexEntry(
                 1324180496,
                 1324180496,
                 64769,
@@ -357,6 +339,7 @@ class IndexEntryFromStatTests(TestCase):
                 12288,
                 "2222222222222222222222222222222222222222",
                 0,
+                None,
             ),
         )
 
